#include <driverlib.h>

#define DATA_SIZE 16*32
#define FREQ_BINS 16
#define BAUDRATE 38400

volatile float cos_table[FREQ_BINS*FREQ_BINS];
volatile float data_in[DATA_SIZE];
volatile float data_out[DATA_SIZE];
volatile float data_out2[DATA_SIZE];
volatile float* data_pointer;
int cos_index, data_index;
String inputString = "";
boolean stringComplete = false;
volatile int num_freq, num_freq2;
int set_clock, set_timer;
int timer_ms = 23;
volatile int timeout = 0;
volatile int clock_freq = 25000;
volatile int timer_counter = 0;
  
void setup()
{
  P2DIR = BIT2 | BIT0;
  P2SEL = BIT2; // Dump SMCLK to P2.2
  P2OUT = 0;
  Serial.begin(BAUDRATE);
  cos_index = 0;
  data_index = 0;
  num_freq = 0;
  set_clock = 0;
  set_timer = 0;
  
  clock_freq = 25000;
  setClock(clock_freq);
  
  reset_blinker();
}

void loop()
{
  if(stringComplete)
    process_input();
}

void serialEvent(){
  while(stringComplete);
  while(Serial.available()){
    char inChar = (char)Serial.read();
    if (inChar != '\n')
      inputString += inChar;
    else
      stringComplete = true;
  }
}

void process_input(){
  if (inputString == "cos_table"){
    cos_index = 1;
    digitalWrite(RED_LED, HIGH);
    Serial.println("\nStart receiving cos_table\n");
  } else if (cos_index){
    load_cos_table();
  } else if (inputString == "data_in"){
    data_index = 1;
    digitalWrite(RED_LED, HIGH);
    Serial.println("\nStart receiving data\n");
  } else if (data_index){
    load_data_in();
  } else if (inputString == "set_clock"){
    set_clock = 1;
    digitalWrite(RED_LED, HIGH);
  } else if (set_clock){
    clock_freq = int(atol(inputString.c_str()));
    set_clock = 0;
    digitalWrite(RED_LED, LOW);
  } else if (inputString == "show_timer"){
    setTimer();
  } else if (inputString == "set_timer"){
    set_timer = 1;
    digitalWrite(RED_LED, HIGH);
  } else if (set_timer){
    setNewTimer();
    digitalWrite(RED_LED, LOW);
  }
  inputString = "";
  stringComplete = false;
}

void load_cos_table(){
  cos_table[cos_index-1] = inputString.toFloat();
  cos_index += 1;
  if(cos_index == FREQ_BINS*FREQ_BINS+1){
    cos_index = 0;
    digitalWrite(RED_LED, LOW);
    Serial.println("\nDone receiving cos_table\n\n");
  }
}

void load_data_in(){
  data_in[data_index-1] = inputString.toFloat();
  data_index += 1;
  if(data_index == DATA_SIZE + 1){
    decode();
  }
}

void send_data_out(){
  Serial.println("\nSending data_out\n");
  digitalWrite(RED_LED, HIGH);
  for(int i = 0; i < DATA_SIZE; i++)
    printFloat(data_out[i], 8);
  Serial.println(num_freq2);
  Serial.println("");
  digitalWrite(RED_LED, LOW);
}

void printFloat( float val, byte precision){
 if (val < 0) Serial.print('-');
 Serial.print (int(abs(val)));  //prints the int part
 if( precision > 0) {
   Serial.print(".");     // print the fractional part
   unsigned long frac;
   unsigned long mult = 1;
   byte padding = precision -1;
   while(precision--)
      mult *=10;
      
   if(val >= 0)
     frac = (val - int(val)) * mult;
   else
     frac = (int(val)- val ) * mult;
   unsigned long frac1 = frac;
   while( frac1 /= 10 )
     padding--;
   while(  padding--)
     Serial.print("0");
   Serial.println(frac,DEC) ;
 } 
}

void decode(){
  data_index = 0;
  digitalWrite(RED_LED, LOW);
  num_freq = 0;
  Serial.println("\nDone receiving data_in\n");
  Serial.end();
  setClock(clock_freq);
  idct();
  setClock(25000);
  Serial.begin(BAUDRATE);
  reconnect();
  send_data_out();
}

void idct(){
  timeout = 0;
  data_pointer = data_out;
  memset((void*)data_out,0,sizeof(data_out));
  int t;
  setTimer();
  while(!timeout){
    for(int f = 0; f < FREQ_BINS; f++){
      for(int n = 0; n < DATA_SIZE; n++){
        t = n & 0xf;
        data_pointer[n] += data_in[n-t+f]*cos_table[t*FREQ_BINS+f];
        if(timeout) break;
      }
      num_freq += 1;
      if(timeout) break;
    }
  }
}

void reconnect(){
  while((char)Serial.read() != 'B') {
    Serial.println('A'); // Reestablish serial  
    delay(100);
  }
}

// Set new MCLK & SMCLK; freq in kHz
void setClock(int freq){
  
  // Set core voltage appropriately
  if (freq < 8000) PMM_setVCore(PMM_CORE_LEVEL_0);
  else if (freq < 12000) PMM_setVCore(PMM_CORE_LEVEL_1);
  else if (freq < 20000) PMM_setVCore(PMM_CORE_LEVEL_2);
  else PMM_setVCore(PMM_CORE_LEVEL_3);
  
  // Set FLL to desired clock frequency
  // Function modified to support Energia's compiler intrisics
  UCS_initFLLSettle2(uint16_t(freq), uint16_t(freq*(1000.0/32768.0)));
  
  // Make sure MCLK and SMCLK are sourced correctly
  if (freq > 16000){
    UCS_clockSignalInit(UCS_MCLK, UCS_DCOCLK_SELECT, UCS_CLOCK_DIVIDER_1);
    UCS_clockSignalInit(UCS_SMCLK, UCS_DCOCLK_SELECT, UCS_CLOCK_DIVIDER_1);
  } else {
    UCS_clockSignalInit(UCS_MCLK, UCS_DCOCLKDIV_SELECT, UCS_CLOCK_DIVIDER_1);
    UCS_clockSignalInit(UCS_SMCLK, UCS_DCOCLKDIV_SELECT, UCS_CLOCK_DIVIDER_1);
  }
  
  set_clock = 0;
  digitalWrite(RED_LED, LOW);
}

void setNewTimer(){
  timer_ms = inputString.toInt();
  set_timer = 0;
}

void setTimer(){
  P2OUT ^= BIT0;
  TA0CCR0 = (unsigned int)(32.768*timer_ms);       // set the timer based on 32kHz clock
  TA0CCTL0 = CCIE;             // enable interrupts for Timer A
  __bis_SR_register(GIE);
  digitalWrite(GREEN_LED, HIGH);
  TA0CTL = TASSEL_1 + MC_1 + TACLR + ID_0;
}

#pragma vector=TIMER0_A0_VECTOR    // Timer A ISR
__interrupt void Timer1_A0_ISR( void )
{
  P2OUT ^= BIT0;
  TA1CTL = MC_0;       // halt timer
  TA0CCTL0 = 0;
  __bic_SR_register(GIE);
  num_freq2 = num_freq;
  data_pointer = data_out2;
  digitalWrite(GREEN_LED, LOW);
  timer_counter++;
  if (timer_counter == 100){
    timer_counter = 0;
    timeout = 1;
  } else 
    setTimer();
}

void reset_blinker(){
  pinMode(RED_LED, OUTPUT);
  pinMode(GREEN_LED, OUTPUT);
  digitalWrite(RED_LED, HIGH);
  delay(100);
  digitalWrite(RED_LED, LOW);
  digitalWrite(GREEN_LED, HIGH);
  delay(100);
  digitalWrite(RED_LED, HIGH);
  digitalWrite(GREEN_LED, LOW);
  delay(100);
  digitalWrite(RED_LED, LOW);
  digitalWrite(GREEN_LED, HIGH);
  delay(100);
  digitalWrite(GREEN_LED, LOW);
}

// The following functions are modified from ucs.c to accomodate Energia's compiler intrinsics

void UCS_initFLLSettle2(uint16_t fsystem, uint16_t ratio){
  volatile uint16_t x = ratio * 32;
  UCS_initFLL2(fsystem, ratio);
  while (x--) __delay_cycles(30);
}

void UCS_initFLL2(uint16_t fsystem, uint16_t ratio){
  uint16_t d, dco_div_bits;
  uint16_t mode = 0;

  //Save actual state of FLL loop control, then disable it. This is needed to
  //prevent the FLL from acting as we are making fundamental modifications to
  //the clock setup.
  uint16_t srRegisterState = __read_status_register() & SCG0;

  d = ratio;
  //Have at least a divider of 2
  dco_div_bits = FLLD__2;

  if (fsystem > 16000) {
    d >>= 1;
    mode = 1;
  } else
    fsystem <<= 1;

  while (d > 512) {
    //Set next higher div level
    dco_div_bits = dco_div_bits + FLLD0;
    d >>= 1;
  }

  // Disable FLL
  __bis_SR_register(SCG0);

  //Set DCO to lowest Tap
  UCSCTL0 = 0x0000;

  //Reset FN bits
  UCSCTL2 &= ~(0x03FF);
  UCSCTL2 = dco_div_bits | (d - 1);

  if (fsystem <= 630)             //fsystem < 0.63MHz
    UCSCTL1 = DCORSEL_0;
  else if (fsystem <  1250)       //0.63MHz < fsystem < 1.25MHz
    UCSCTL1 = DCORSEL_1;
  else if (fsystem <  2500)       //1.25MHz < fsystem <  2.5MHz
    UCSCTL1 = DCORSEL_2;
  else if (fsystem <  5000)       //2.5MHz  < fsystem <    5MHz
    UCSCTL1 = DCORSEL_3;
  else if (fsystem <  10000)      //5MHz    < fsystem <   10MHz
    UCSCTL1 = DCORSEL_4;
  else if (fsystem <  20000)      //10MHz   < fsystem <   20MHz
    UCSCTL1 = DCORSEL_5;
  else if (fsystem <  40000)      //20MHz   < fsystem <   40MHz
    UCSCTL1 = DCORSEL_6;
  else
    UCSCTL1 = DCORSEL_7;

  // Re-enable FLL
  __bic_SR_register(SCG0);

  while (UCSCTL7 & DCOFFG) {
    //Clear OSC flaut Flags
    UCSCTL7 &= ~(DCOFFG);

    //Clear OFIFG fault flag
    SFRIFG1 &= ~OFIFG;
  }

  // Restore previous SCG0
  __bis_SR_register(srRegisterState);

  if (mode == 1) {
    //fsystem > 16000
    //Select DCOCLK
    UCSCTL4 &=  ~(SELM_7 + SELS_7);
    UCSCTL4 |= SELM__DCOCLK + SELS__DCOCLK;
  } else {
    //Select DCODIVCLK
    UCSCTL4 &=  ~(SELM_7 + SELS_7);
    UCSCTL4 |= SELM__DCOCLKDIV + SELS__DCOCLKDIV;
  }
}
